@@ -6,6 +6,7 @@ from tuneflow_py.models.clip import Clip, ClipType
 from tuneflow_py.models.tempo import TempoEvent
 from tuneflow_py.models.time_signature import TimeSignatureEvent
 from tuneflow_py.models.automation import AutomationTarget, AutomationTargetType
+from tuneflow_py.models.audio_plugin import AudioPlugin, decode_audio_plugin_tuneflow_id
 from tuneflow_py.utils import db_to_volume_value, greater_equal, lower_than
 from miditoolkit.midi import MidiFile, TempoChange as ToolkitTempoChange, TimeSignature as ToolkitTimeSignature, Instrument, Note as ToolkitNote
 from types import SimpleNamespace
@@ -19,10 +20,8 @@ class Song:
         else:
             self._proto = song_pb2.Song()
             self._proto.PPQ = Song.get_default_resolution()
-            self._proto.tempos.append(
-                song_pb2.TempoEvent(ticks=0, time=0, bpm=120))
-            self._proto.time_signatures.append(
-                song_pb2.TimeSignatureEvent(ticks=0, numerator=4, denominator=4))
+            self._proto.tempos.add(ticks=0, time=0, bpm=120)
+            self._proto.time_signatures.add(ticks=0, numerator=4, denominator=4)
             self._proto.last_tick = 0
             self._proto.duration = 0
             self._proto.master_track.uuid = Track._generate_track_id()
@@ -81,8 +80,12 @@ class Song:
         # Add tempos and time signatures
         song.overwrite_tempo_changes([TempoEvent(ticks=scale_int_by(
             tempo_change.time, ppq_scale_factor), bpm=tempo_change.tempo) for tempo_change in midi_obj.tempo_changes])
-        song.overwrite_time_signature_changes([TimeSignatureEvent(ticks=scale_int_by(
-            time_signature_change.time, ppq_scale_factor), numerator=time_signature_change.numerator, denominator=time_signature_change.denominator) for time_signature_change in midi_obj.time_signature_changes])
+        song.overwrite_time_signature_changes([
+            TimeSignatureEvent(
+                ticks=scale_int_by(time_signature_change.time, ppq_scale_factor),
+                numerator=time_signature_change.numerator,
+                denominator=time_signature_change.denominator)
+            for time_signature_change in midi_obj.time_signature_changes])
 
         # Add tracks and notes.
         song_last_tick = 0
@@ -98,8 +101,10 @@ class Song:
                 start_tick = scale_int_by(
                     note.start, ppq_scale_factor)
                 end_tick = scale_int_by(note.end, ppq_scale_factor)
-                track_clip_proto.notes.add(pitch=note.pitch, velocity=note.velocity, start_tick=start_tick, start_time=song.tick_to_seconds(
-                    start_tick), end_tick=end_tick, end_time=song.tick_to_seconds(end_tick))
+                track_clip_proto.notes.add(
+                    pitch=note.pitch, velocity=note.velocity, start_tick=start_tick, start_time=song.tick_to_seconds(
+                        start_tick),
+                    end_tick=end_tick, end_time=song.tick_to_seconds(end_tick))
             track_clip_proto.clip_start_tick = min(
                 track_clip_proto.notes, key=lambda x: x.start_tick).start_tick
             track_clip_proto.clip_end_tick = max(
@@ -123,6 +128,7 @@ class Song:
                 volume_target_id = volume_target.to_tf_automation_target_id()
                 song_track_proto.automation.targets.append(
                     volume_target._proto)
+                volume_target._proto = song_track_proto.automation.targets[-1]
                 volume_target_value = song_pb2.AutomationValue()
                 song_track_proto.automation.target_values[volume_target_id] = volume_target_value
                 for index, cc in enumerate(sorted(volume_ccs, key=lambda x: x.time)):
@@ -138,6 +144,7 @@ class Song:
                 pan_target = AutomationTarget(AutomationTargetType.PAN)
                 pan_target_id = pan_target.to_tf_automation_target_id()
                 song_track_proto.automation.targets.append(pan_target._proto)
+                pan_target._proto = song_track_proto.automation.targets[-1]
                 pan_target_value = song_pb2.AutomationValue()
                 song_track_proto.automation.target_values[pan_target_id] = pan_target_value
                 for index, cc in enumerate(sorted(pan_ccs, key=lambda x: x.time)):
@@ -158,8 +165,11 @@ class Song:
             midi_obj.tempo_changes.append(ToolkitTempoChange(
                 tempo=tempo_proto.bpm, time=tempo_proto.ticks))
         for time_signature_proto in self._proto.time_signatures:
-            midi_obj.time_signature_changes.append(ToolkitTimeSignature(
-                numerator=time_signature_proto.numerator, denominator=time_signature_proto.denominator, time=time_signature_proto.ticks))
+            midi_obj.time_signature_changes.append(
+                ToolkitTimeSignature(
+                    numerator=time_signature_proto.numerator,
+                    denominator=time_signature_proto.denominator,
+                    time=time_signature_proto.ticks))
         for track_proto in self._proto.tracks:
             if track_proto.type != TrackType.MIDI_TRACK or len(track_proto.clips) == 0:
                 continue
@@ -171,10 +181,14 @@ class Song:
                 if clip_proto.type != ClipType.MIDI_CLIP:
                     continue
                 for note_proto in clip_proto.notes:
-                    if not Clip.is_note_in_clip(note_start_tick=note_proto.start_tick, note_end_tick=note_proto.end_tick, clip_start_tick=clip_proto.clip_start_tick, clip_end_tick=clip_proto.clip_end_tick):
+                    if not Clip.is_note_in_clip(
+                            note_start_tick=note_proto.start_tick, note_end_tick=note_proto.end_tick,
+                            clip_start_tick=clip_proto.clip_start_tick, clip_end_tick=clip_proto.clip_end_tick):
                         continue
-                    instrument.notes.append(ToolkitNote(
-                        pitch=note_proto.pitch, velocity=note_proto.velocity, start=note_proto.start_tick, end=note_proto.end_tick))
+                    instrument.notes.append(
+                        ToolkitNote(
+                            pitch=note_proto.pitch, velocity=note_proto.velocity,
+                            start=note_proto.start_tick, end=note_proto.end_tick))
             # TODO: Export automation
         return midi_obj
 
@@ -208,8 +222,10 @@ class Song:
         )
         if insert_index < 0:
             self._proto.tempos.append(tempo_change._proto)
+            tempo_change._proto = self._proto.tempos[-1]
         else:
             self._proto.tempos.insert(insert_index, tempo_change._proto)
+            tempo_change._proto = self._proto.tempos[insert_index]
 
         self.retiming_tempo_events()
         return tempo_change
@@ -247,6 +263,29 @@ class Song:
         )
         return base_tempo_change.time + ticks_delta / ticks_per_second_since_last_tempo_change
 
+    def seconds_to_tick(self, seconds: float):
+        if (seconds == 0):
+            return 0
+
+        target_tempo = SimpleNamespace()
+        target_tempo.time = seconds
+        base_tempo_index = lower_than(
+            self._proto.tempos,
+            target_tempo,
+            key=lambda x: x.time
+        )
+        if (base_tempo_index == -1):
+            # If no tempo is found before the time, use the first tempo.
+            base_tempo_index = 0
+
+        base_tempo_change_proto = self._proto.tempos[base_tempo_index]
+        time_delta = seconds - base_tempo_change_proto.time
+        ticks_per_second_since_last_tempo_change = Song._tempo_bpm_to_ticks_per_second(
+            base_tempo_change_proto.bpm,
+            self.get_resolution(),
+        )
+        return round(base_tempo_change_proto.ticks + time_delta * ticks_per_second_since_last_tempo_change)
+
     def overwrite_tempo_changes(self, tempo_events: List[TempoEvent]):
         if len(tempo_events) == 0:
             raise Exception('Cannot clear all the tempo events.')
@@ -256,8 +295,8 @@ class Song:
         if first_tempo_event.get_ticks() > 0:
             raise Exception('The first tempo event needs to start from tick 0')
         del self._proto.tempos[:]
-        self._proto.tempos.append(song_pb2.TempoEvent(
-            ticks=0, time=0, bpm=first_tempo_event.get_bpm()))
+        self._proto.tempos.add(
+            ticks=0, time=0, bpm=first_tempo_event.get_bpm())
         for i in range(1, len(sorted_tempo_events)):
             tempo_event = sorted_tempo_events[i]
             self.create_tempo_change(
@@ -270,7 +309,9 @@ class Song:
         del self._proto.time_signatures[:]
         for time_signature_change in time_signatures:
             self._proto.time_signatures.add(
-                ticks=time_signature_change.get_ticks(), numerator=time_signature_change.get_numerator(), denominator=time_signature_change.get_denominator())
+                ticks=time_signature_change.get_ticks(),
+                numerator=time_signature_change.get_numerator(),
+                denominator=time_signature_change.get_denominator())
 
     def get_time_signature_event_count(self):
         return len(self._proto.time_signatures)
@@ -278,17 +319,48 @@ class Song:
     def get_time_signature_event_at(self, index: int):
         return TimeSignatureEvent(proto=self._proto.time_signatures[index])
 
+    def create_audio_plugin(self, tf_id: str):
+        pluginInfo = decode_audio_plugin_tuneflow_id(tf_id)
+        plugin = AudioPlugin(
+            name=pluginInfo["name"],
+            manufacturer_name=pluginInfo["manufacturer_name"],
+            plugin_format_name=pluginInfo["plugin_format_name"],
+            plugin_version=pluginInfo["plugin_version"],
+        )
+        return plugin
+
     def create_track(self, type: int,
-                     index: int,
-                     rank: int | None = None):
+                     index: int | None = None,
+                     rank: int | None = None,
+                     assign_default_sampler_plugin=False):
+        '''
+        Adds a new track into the song and returns it.
+
+        @param index Index to insert at. If left blank, appends to the end.
+        @param rank The displayed rank which uniquely identifies a track. Internal use, do not set this.
+        @param assign_default_sampler_plugin Whether to assign a default sampler plugin if type is `MIDI_TRACK`.
+        '''
         new_track = Track(
             type=type, song=self, rank=rank if rank is not None else self.get_next_track_rank())
+        if assign_default_sampler_plugin and type == TrackType.MIDI_TRACK:
+            new_track.set_sampler_plugin(new_track.create_audio_plugin(AudioPlugin.DEFAULT_SYNTH_TF_ID))
+        if type == TrackType.AUX_TRACK:
+            # TODO: Set default input bus rank.
+            pass
+            # new_track.getAuxTrackData().setInputBusRank(1)
+
+        if index is None:
+            index = len(self._proto.tracks)
         self._proto.tracks.insert(index, new_track._proto)
+        new_track._proto = self._proto.tracks[index]
         return new_track
 
     def get_next_track_rank(self):
         return 1 if len(self._proto.tracks) == 0 else max([track.rank for track in self._proto.tracks]) + 1
 
+    def __repr__(self) -> str:
+        return str(self._proto)
+
     @staticmethod
     def _tempo_bpm_to_ticks_per_second(tempo_bpm: float, PPQ: int):
         return (tempo_bpm * PPQ) / 60
